package java.util.concurrent;

import java.util.function.Function;
import java.util.function.BiFunction;
import java.util.Iterator;
import java.util.Objects;
import java.util.function.BiConsumer;
import java.util.Map;

public interface ConcurrentMap<K, V> extends Map<K, V>
{
    default public V getOrDefault(Object o, V v) {
        Object value;
        return ((value = this.get(o)) != null) ? value : v;
    }
    
    default public void forEach(BiConsumer<? super K, ? super V> obj) {
        Objects.requireNonNull(obj);
        for (Map.Entry entry : this.entrySet()) {
            Object key;
            Object value;
            try {
                key = entry.getKey();
                value = entry.getValue();
            }
            catch (IllegalStateException ex) {
                continue;
            }
            obj.accept(key, value);
        }
    }
    
    public abstract V putIfAbsent(K p0, V p1);
    
    public abstract boolean remove(Object p0, Object p1);
    
    public abstract boolean replace(K p0, V p1, V p2);
    
    public abstract V replace(K p0, V p1);
    
    default public void replaceAll(BiFunction<? super K, ? super V, ? extends V> obj) {
        Objects.requireNonNull(obj);
        this.forEach((o, value)/* java.lang.Object,java.lang.Object, */ -> {
            while (!this.replace(o, value, p0.apply(o, value)) && (value = this.get(o)) != null) {}
            return;
        });
    }
    
    default public V computeIfAbsent(K k, Function<? super K, ? extends V> obj) {
        Objects.requireNonNull(obj);
        Object o;
        Object apply;
        return ((o = this.get(k)) == null && (apply = obj.apply(k)) != null && (o = this.putIfAbsent(k, apply)) == null) ? apply : o;
    }
    
    default public V computeIfPresent(K k, BiFunction<? super K, ? super V, ? extends V> obj) {
        Objects.requireNonNull(obj);
        Object value;
        while ((value = this.get(k)) != null) {
            Object apply = obj.apply(k, value);
            if (apply != null) {
                if (this.replace(k, value, apply)) {
                    return apply;
                }
                continue;
            }
            else {
                if (this.remove(k, value)) {
                    return null;
                }
                continue;
            }
        }
        return value;
    }
    
    default public V compute(K k, BiFunction<? super K, ? super V, ? extends V> obj) {
        Objects.requireNonNull(obj);
        Object o = this.get(k);
        while (true) {
            Object apply = obj.apply(k, o);
            if (apply == null) {
                if (o != null || this.containsKey(k)) {
                    if (this.remove(k, o)) {
                        return null;
                    }
                    o = this.get(k);
                }
                else {
                    return null;
                }
            }
            else if (o != null) {
                if (this.replace(k, o, apply)) {
                    return apply;
                }
                o = this.get(k);
            }
            else {
                if ((o = this.putIfAbsent(k, apply)) == null) {
                    return apply;
                }
                continue;
            }
        }
    }
    
    default public V merge(K k, V obj, BiFunction<? super V, ? super V, ? extends V> obj2) {
        Objects.requireNonNull(obj2);
        Objects.requireNonNull(obj);
        Object o = this.get(k);
        while (true) {
            if (o != null) {
                Object apply = obj2.apply(o, obj);
                if (apply != null) {
                    if (this.replace(k, o, apply)) {
                        return apply;
                    }
                }
                else if (this.remove(k, o)) {
                    return null;
                }
                o = this.get(k);
            }
            else {
                if ((o = this.putIfAbsent(k, obj)) != null) {
                    continue;
                }
                return obj;
            }
        }
    }
}
